// copyright 2002 Josh Levenberg, please see copying.txt

/* see disp_nv.cpp for comments */
#ifndef NVIDIA_DISPLAY_H
#define NVIDIA_DISPLAY_H
#include "display.h"
#include <gl/glut.h>
#include <jstring.h>
#include <deque>
#include <vector>

class NvidiaDisplay : public AbstractDisplay
  {
public:
    NvidiaDisplay();

    void Initialize(int maxnumvert, int maxnumind);
    void Done();

    // Cached geometry
    bool BeginCache(Coord *&, Index *&);
    Cache FinishCache(int numvert, int numind);
    void RenderCache(Cache);
    void DiscardCache(Cache);

protected:
    enum { SHORT_INDICES=1 };
    typedef unsigned short indextype;
    enum { indexenum = GL_UNSIGNED_SHORT };
    // 65536 is the max index for nVidia GeForce Processors
    enum { MaxIndex=65536 };

    struct CacheData
      {
        indextype *mTrisBegin;
        indextype *mTrisEnd;
        union
          {
            int Next;          // if not active
            int LastRender;    // if active, mFenceGroups[u.LastRender-mFirstFenceGroup]
          } u;
#ifdef RANGE_ELEMENTS_NV
        int mNumVerts;
#endif
#ifndef NDEBUG
        bool mDiscarded;
#endif
      };

    struct FenceGroup
      {
        GLuint mFence;
        int mFirstPending;
        int mLastPending;
      };

    int mFirstFenceGroup;
    std::deque<FenceGroup> mFenceGroups;
    int mNumLeftLastGroup;
    int mCachesPerGroup;

    std::vector<CacheData> mCaches;
    int mFirstFreeCache;
    // mBlocksUsed replaced by mCaches.size()

    int mBlockSize;
    int mNumBlocks;
    float *mVideoMemory;
    // mVideoMemory has mBlockSize*mNumBlocks*FloatsPerVert*sizeof(float)

    int mDefaultTriMemory;
    int mCurrentCache;

    static void SetVidMemOffsetPtr(float *ptr)
      {
        glVertexPointer(3, GL_FLOAT, FloatsPerVert*sizeof(float), ptr);
        glTexCoordPointer(2, GL_FLOAT, FloatsPerVert*sizeof(float), ptr);
      }
    void PopGroup();
    void AddToLastGroup(int cache);
    void CloseLastGroup();

#ifndef NDEBUG
    void CheckBlockConsistency();
#endif
  };

#endif
