// copyright 2002 Josh Levenberg, please see copying.txt

/* see disp_va.cpp for comments */

#ifndef VERTEX_ARRAY_DISPLAY_H
#define VERTEX_ARRAY_DISPLAY_H
#include "display.h"
#include <gl/glut.h>

// If defined, uses glDrawRangeElements instead of glDrawElements
//#define RANGE_ELEMENTS_VA

class VertexArrayDisplay : public AbstractDisplay
  {
public:
    VertexArrayDisplay() 
      : mVertices(0), mIndices(0),
        mFirstCache(0), mFreeCache(0), mCurrentCache(0) 
      { }

    void Initialize(int maxnumvert, int maxnumind);
    void Done();

    // Cached geometry
    bool BeginCache(Coord *&, Index *&);
    Cache FinishCache(int numvert, int numind);
    void RenderCache(Cache);
    void DiscardCache(Cache);

protected:
    enum { SHORT_INDICES=1 };
    typedef unsigned short indextype;
    enum { indexenum = GL_UNSIGNED_SHORT };

    int mBlockSize;
    int mDefaultTriMemory;

    struct cachestruct
      {
        float *mVertices;
        indextype *mIndices;
#ifdef RANGE_ELEMENTS_VA
        int mNumVertices;
#endif
        int mNumIndices;
        cachestruct *mNext;
        cachestruct *mPrev;
      };
    typedef cachestruct * CacheType;
    CacheType mFirstCache;
    CacheType mFreeCache;
    CacheType mCurrentCache;
    static inline bool IsValidCache(CacheType c) 
      { return c!=0 && c!=reinterpret_cast<CacheType>(-1); }
    void CheckBlockConsistency();

    float *mVertices;
    indextype *mIndices;
  };
#endif