#include "dlod.h"
#include "dispmeth.h"
#include "roamstd.h"
#include "newalgo.h"
#include "display.h"
#include "glext.h"
#include "uitype.h"

// Display method switching

int NumDisplayMethods;
DisplayMethodType gDisplayMethods[MaxDisplayMethods];
int gCurrentDisplay = -1;

static void InitDisplayNewAlgo_Raw()
  {
    fprintf(gOutFile, "Caching disabled\n");
    printf("Caching disabled\n");
    SelectVertexArrayDisplay();
    InitDisplayNewAlgo();
  }

static void InitDisplayNewAlgo_NV()
  {
    fprintf(gOutFile, "nVidia Caching enabled\n");
    printf("nVidia Caching enabled\n");
    SelectNVDisplay();
    InitDisplayNewAlgo();
  }

static void InitDisplayFullMesh_Raw()
  {
    fprintf(gOutFile, "Caching disabled\n");
    printf("Caching disabled\n");
    SelectVertexArrayDisplay();
    InitDisplayFullMesh();
  }

static void InitDisplayFullMesh_NV()
  {
    fprintf(gOutFile, "nVidia Caching enabled\n");
    printf("nVidia Caching enabled\n");
    SelectNVDisplay();
    InitDisplayFullMesh();
  }

void SwitchToDisplay(int newdisplay)
  {
    if (newdisplay!=gCurrentDisplay)
      {
        if (gCurrentDisplay>=0 && gCurrentDisplay<NumDisplayMethods)
          gDisplayMethods[gCurrentDisplay].close();
        gCurrentDisplay=newdisplay;
        if (gCurrentDisplay>=0 && gCurrentDisplay<NumDisplayMethods)
          {
            glutKeyboardFunc(gDisplayMethods[gCurrentDisplay].keyboard);
            gDisplayMethods[gCurrentDisplay].init();
          }
        else
          {
            glutKeyboardFunc(DefaultKeyboard);
          }
      }
  }

void InitDisplayMethods()
  {
    NumDisplayMethods=0;

#ifndef INTERACTIVE_MODE
    gDisplayMethods[NumDisplayMethods].init=InitDisplayNewAlgo_Raw;
    gDisplayMethods[NumDisplayMethods].draw=DrawNewAlgo;
    gDisplayMethods[NumDisplayMethods].close=CloseDisplayNewAlgo;
    gDisplayMethods[NumDisplayMethods].keyboard=KeyboardNewAlgo;
    gDisplayMethods[NumDisplayMethods].printstats=PrintStatsNewAlgo;
    gDisplayMethods[NumDisplayMethods].resetstats=ResetStatsNewAlgo;
    ++NumDisplayMethods;
#endif

    gDisplayMethods[NumDisplayMethods].init=InitDisplayRoamStandard;
    gDisplayMethods[NumDisplayMethods].draw=DrawRoamStandard;
    gDisplayMethods[NumDisplayMethods].close=CloseDisplayRoamStandard;
    gDisplayMethods[NumDisplayMethods].keyboard=DefaultKeyboard;
    gDisplayMethods[NumDisplayMethods].printstats=PrintStatsRoamStandard;
    gDisplayMethods[NumDisplayMethods].resetstats=ResetStatsRoamStandard;
    ++NumDisplayMethods;

#ifndef INTERACTIVE_MODE
    gDisplayMethods[NumDisplayMethods].init=InitDisplayFullMesh_Raw;
    gDisplayMethods[NumDisplayMethods].draw=DrawFullMesh;
    gDisplayMethods[NumDisplayMethods].close=CloseDisplayFullMesh;
    gDisplayMethods[NumDisplayMethods].keyboard=DefaultKeyboard;
    gDisplayMethods[NumDisplayMethods].printstats=PrintStatsFullMesh;
    gDisplayMethods[NumDisplayMethods].resetstats=ResetStatsFullMesh;
    ++NumDisplayMethods;
#endif

#ifndef MOVIE_OUTPUT
    if (gNVExtensions)
      {
        /* FullMesh_NV runs out of video memory at HeightFieldTesselation >= 9 */
#ifndef INTERACTIVE_MODE
        gDisplayMethods[NumDisplayMethods].init=InitDisplayFullMesh_NV;
        gDisplayMethods[NumDisplayMethods].draw=DrawFullMesh;
        gDisplayMethods[NumDisplayMethods].close=CloseDisplayFullMesh;
        gDisplayMethods[NumDisplayMethods].keyboard=DefaultKeyboard;
        gDisplayMethods[NumDisplayMethods].printstats=PrintStatsFullMesh;
        gDisplayMethods[NumDisplayMethods].resetstats=ResetStatsFullMesh;
        ++NumDisplayMethods;
#endif

        gDisplayMethods[NumDisplayMethods].init=InitDisplayNewAlgo_NV;
        gDisplayMethods[NumDisplayMethods].draw=DrawNewAlgo;
        gDisplayMethods[NumDisplayMethods].close=CloseDisplayNewAlgo;
        gDisplayMethods[NumDisplayMethods].keyboard=KeyboardNewAlgo;
        gDisplayMethods[NumDisplayMethods].printstats=PrintStatsNewAlgo;
        gDisplayMethods[NumDisplayMethods].resetstats=ResetStatsNewAlgo;
        ++NumDisplayMethods;
        SwitchToDisplay(NumDisplayMethods-1);
      }
    else
#endif
      {
#ifdef INTERACTIVE_MODE
        gDisplayMethods[NumDisplayMethods].init=InitDisplayNewAlgo_Raw;
        gDisplayMethods[NumDisplayMethods].draw=DrawNewAlgo;
        gDisplayMethods[NumDisplayMethods].close=CloseDisplayNewAlgo;
        gDisplayMethods[NumDisplayMethods].keyboard=KeyboardNewAlgo;
        gDisplayMethods[NumDisplayMethods].printstats=PrintStatsNewAlgo;
        gDisplayMethods[NumDisplayMethods].resetstats=ResetStatsNewAlgo;
        ++NumDisplayMethods;
        SwitchToDisplay(NumDisplayMethods-1);
#else
        SwitchToDisplay(0);
#endif
      }
  }