#include "frustum.h"
#include "dlod.h"

struct ClipPlane
  {
    float nx, ny, offs;
  };

static ClipPlane gClipPlanes[3];

void UpdateClipPlanes()
  {
    // read modelview matrix?
    // what about projection matrix?

    // assume can derive planes from:
    // gWidth/gHeight
    // gCamera[0].mFov  // degrees in y direction
    // gCamera[0].mLocation
    // gCamera[0].mLookAt

    // for the moment, only care about left, right, and near planes

    double xtan = tan(gCamera[0].mFov/2)*gWidth/gHeight;
    float vx = gCamera[0].mLookAt[0] - gCamera[0].mLocation[0];
    float vy = gCamera[0].mLookAt[1] - gCamera[0].mLocation[1];
    float tvx = xtan*vx;
    float tvy = xtan*vy;
    /*    
     l    ^    r
      \   |v  /
       \     /
        \   /
         \ /
      ----*---- near
     */
    gClipPlanes[0].nx = -vy-tvx;
    gClipPlanes[0].ny = vx-tvy;
    gClipPlanes[0].offs = -(gCamera[0].mLocation[0]*gClipPlanes[0].nx+
                            gCamera[0].mLocation[1]*gClipPlanes[0].ny);

    gClipPlanes[1].nx = vy-tvx;
    gClipPlanes[1].ny = -vx-tvy;
    gClipPlanes[1].offs = -(gCamera[0].mLocation[0]*gClipPlanes[1].nx+
                            gCamera[0].mLocation[1]*gClipPlanes[1].ny);

    gClipPlanes[2].nx = -vx;
    gClipPlanes[2].ny = -vy;
    gClipPlanes[2].offs = -(gCamera[0].mLocation[0]*gClipPlanes[2].nx+
                            gCamera[0].mLocation[1]*gClipPlanes[2].ny);
  }

static inline bool OutsideCheck(int plane, float x, float y)
  {
    return (x*gClipPlanes[plane].nx + 
            y*gClipPlanes[plane].ny + 
            gClipPlanes[plane].offs > 0);
  }

int IsVisible(int wascrossing, const std::vector<vec3s_t> &bounds)
  { // for each edge of the view cone
    int crossing=0;
    for (int side=0; side<3; ++side)
      if (wascrossing & (1<<side))
        { // if all vertices outside that edge, return false
          int i;
          for (i=0; i<bounds.size(); ++i)
            if (! OutsideCheck(side, bounds[i].m[0], bounds[i].m[1]))
              break;
        if (i==bounds.size())
          return Outside;
        else if (i>0)
          crossing |= 1<<side;
        else
          {
            for (i=1; i<bounds.size(); ++i)
              if (! OutsideCheck(side, bounds[i].m[0], bounds[i].m[1]))
                {
                  crossing |= 1<<side;
                  break;
                }
          }
        }
    if (crossing)
      return crossing;
    else
      return Inside;
  }

//int VisibleTree(int treenum)
//  {
//    static std::vector<vec3s_t> bound;
//    bound.clear();
//    gModel->BoundingVolume(treenum, bound);
//    return IsVisible(bound);
//  }
