// copyright 2002 Josh Levenberg, please see copying.txt
#include "uitype.h"

#ifdef INTERACTIVE_MODE
#include "dlod.h"
#include "model.h"
#include "os.h"

static int sSpeedUp=1;
static double sAvgTimeStep=0;
static bool sUpdate=true;
static float sAvatarHdg = 0;
static float sAvatarVel[3] = {0, 0, 0};
static double sLastTime;

#define AVG_AMOUNT (0.9)
#define HARD_LIMIT (gAvatarHeight/2)

#ifdef AUTOPILOT
static float sOrbitRadius = 0.24;

void FocusChange(int state)
  {
    if (state==GLUT_LEFT)
      gPaused=true;
  }

static vec_t sAngle=0;

void AnimateAvatar(double stepsize)
  { // stepsize should average sSpeedUp
    if (sUpdate || !gPaused)
      {
        if (!gPaused)
          sAngle += stepsize*2*PI/2400;

        gAvatar[0] = 0.5+sOrbitRadius*cos(sAngle);
        gAvatar[1] = 0.5+sOrbitRadius*sin(sAngle);

        sAvatarHdg = sAngle+PI/2;

        sUpdate=true;

        double groundheight=gModel->LookUpHeight(gAvatar[0], gAvatar[1]);
        double idealheight=groundheight+gAvatarHeight;
        double minheight=groundheight+HARD_LIMIT;
        gAvatar[2] += (idealheight-gAvatar[2])*(1-pow(0.93, stepsize));
        if (gAvatar[2] < minheight)
          gAvatar[2]=minheight;
      }
  }

#else

bool sMouseButton[2];
double sMousePos[2];

void MouseClick(int button, int state, int x, int y)
  {
    int butnum = -1;
    if (button==GLUT_LEFT_BUTTON)
      butnum=0;
    else if (button==GLUT_RIGHT_BUTTON)
      butnum=1;
    if (butnum>=0)
      {
        bool statenum = false;
        if (state==GLUT_DOWN)
          statenum=true;
        sMouseButton[butnum]=statenum;
      }
    sMousePos[0]=((double)x)/gWidth;
    sMousePos[1]=((double)y)/gHeight;
  }

void MouseDrag(int x, int y)
  {
    sMousePos[0]=((double)x)/gWidth; 
    sMousePos[1]=((double)y)/gHeight;
  }

void MouseMove(int x, int y)
  {
    sMousePos[0]=((double)x)/gWidth;
    sMousePos[1]=((double)y)/gHeight;
  }

#define STEERING_CENTER (0.05)
#define STEERING_LEFT (.5-STEERING_CENTER)
#define STEERING_RIGHT (.5+STEERING_CENTER)
#define STEERING_SPEED (PI/60)
#define RUNNING_SPEED (1./240)
#define FRICTION (1./1200)

double SteeringValue(double mouse)
  {
    double steering = 0;
    if (mouse<STEERING_LEFT)
      {
        steering = mouse-STEERING_LEFT;
        steering *= steering;
        steering = -steering;
        steering /= STEERING_LEFT*STEERING_LEFT;
        return steering;
      }
    else if (mouse>STEERING_RIGHT)
      {
        steering = mouse-STEERING_RIGHT;
        steering *= steering;
        steering /= STEERING_LEFT*STEERING_LEFT;
      }
    return steering;
  }

void AnimateAvatar(double stepsize)
  { // stepsize should average sSpeedUp
    if (!gPaused)
      {
        // sAvatarVel
        vec3_t oldpos;
        oldpos[0]=gAvatar[0];
        oldpos[1]=gAvatar[1];
        oldpos[2]=gAvatar[2];
        if (sMouseButton[0])
          {
            sUpdate=true;
            //sAvatarVel
            sAvatarHdg -= STEERING_SPEED*SteeringValue(sMousePos[0])*stepsize;
            double velocity = -RUNNING_SPEED*SteeringValue(sMousePos[1])*stepsize;
            gAvatar[0] += velocity*cos(sAvatarHdg);
            gAvatar[1] += velocity*sin(sAvatarHdg);
          }
        else
          {
            double velocity=sqrt(vdot(sAvatarVel, sAvatarVel));
            if (velocity > 0.00000001)
              {
                double slowdown=stepsize*FRICTION;
                sUpdate=true;
                if (velocity > slowdown)
                  {
                    slowdown = (velocity-slowdown)/velocity;
                    gAvatar[0] += sAvatarVel[0]*slowdown;
                    gAvatar[1] += sAvatarVel[1]*slowdown;
                  }
              }
          }
        if (gAvatar[0]<0)
          gAvatar[0]=0;
        else if (gAvatar[0]>1)
          gAvatar[0]=1;
        if (gAvatar[1]<0)
          gAvatar[1]=0;
        else if (gAvatar[1]>1)
          gAvatar[1]=1;
        double groundheight=gModel->LookUpHeight(gAvatar[0], gAvatar[1]);
        double idealheight=groundheight+gAvatarHeight;
        if (sUpdate)
          {
            double minheight=groundheight+HARD_LIMIT;
            gAvatar[2] += (idealheight-gAvatar[2])*(1-pow(0.93, stepsize));
            if (gAvatar[2] < minheight)
              gAvatar[2]=minheight;
          }
        else
          gAvatar[2] = idealheight;
        sAvatarVel[0]=(gAvatar[0]-oldpos[0])/stepsize;
        sAvatarVel[1]=(gAvatar[1]-oldpos[1])/stepsize;
        sAvatarVel[2]=(gAvatar[2]-oldpos[2])/stepsize;
      }
  }

#endif

void glutIdle()
  { 
    double time = GetTimeInMS()/1000.0;
    double dt = time - sLastTime;
    sLastTime = time;
    double step;
    if (sAvgTimeStep==0)
      {
        if (!gPaused)
          sAvgTimeStep=dt;
        step=1;
      }
    else 
      {
        if (dt > 2.0*sAvgTimeStep)
          dt=2.0*sAvgTimeStep;
        else if (dt < 0.5*sAvgTimeStep)
          dt=0.5*sAvgTimeStep;
        step = dt / sAvgTimeStep;
        if (!gPaused)
          {
            sAvgTimeStep *= AVG_AMOUNT;
            sAvgTimeStep += (1-AVG_AMOUNT)*dt;
          }
      }
    
    AnimateAvatar(step*sSpeedUp);

    if (sUpdate)
      {
        gCamera[0].mLocation[0] = gAvatar[0];
        gCamera[0].mLocation[1] = gAvatar[1];
        gCamera[0].mLocation[2] = gAvatar[2];

        gCamera[0].mLookAt[0] = gAvatar[0] + cos(sAvatarHdg)*0.8;
        gCamera[0].mLookAt[1] = gAvatar[1] + sin(sAvatarHdg)*0.8;
        gCamera[0].mLookAt[2] = gAvatar[2];

        gCamera[1].mLocation[0] = gCamera[0].mLookAt[0];
        gCamera[1].mLocation[1] = gCamera[0].mLookAt[1];

        gCamera[1].mLookAt[0] = gCamera[1].mLocation[0];
        gCamera[1].mLookAt[1] = gCamera[1].mLocation[1];
        gCamera[1].mLookAt[2] = gAvatar[2];

        glutPostRedisplay();
        sUpdate=false;
      }
    else
      {
        Sleep(50); // 50 ms
      }
  }

void UIInit()
  {
#if defined(AUTOPILOT)
    glutVisibilityFunc(FocusChange);
#else
    glutMouseFunc(MouseClick);
    glutMotionFunc(MouseDrag);
    glutPassiveMotionFunc(MouseMove);
#endif
    sLastTime = GetTimeInMS()/1000.0;
  }

void UIKey(unsigned char key)
  {
    switch (key)
      {
        case 'a':
        case 'A':
        case 'z':
        case 'Z':
        case '+':
        case '-':
        case '1':
        case '2':
        case '3':
        case 'w':
        case 'W':
        case 'f':
        case 'F':
        case 'd':
        case 'D':
        case 'c':
        case 'C':
          sUpdate=true;
          break;

#if defined(AUTOPILOT)
        case ' ':
          // pause/unpause
          gPaused = !gPaused;
          break;
#endif

        case '\'':
          sSpeedUp *= 2;
          fprintf(gOutFile, "SpeedUp doubled to: %d\n", sSpeedUp);
          break;

        case ';':
          if (sSpeedUp>1)
            {
              sSpeedUp /= 2;
              fprintf(gOutFile, "SpeedUp halved to: %d\n", sSpeedUp);
            }
          break;
      }
  }

void UIPostDraw()
  {
  }
#endif